package de.fraunhofer.iem.dataprovider.kpi.metrics

import de.fraunhofer.iem.dataprovider.kpi.metrics.repositoryDetails.service.RepositoryDetailsService
import de.fraunhofer.iem.dataprovider.kpi.service.KPIService
import de.fraunhofer.iem.dataprovider.repository.service.RepositoryService
import de.fraunhofer.iem.dataprovider.toolRun.service.ToolRunService
import de.fraunhofer.iem.dataprovider.tools.occmd.service.OccmdService
import de.fraunhofer.iem.dataprovider.tools.ort.service.OrtService
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.async
import kotlinx.coroutines.launch
import org.springframework.stereotype.Service

@Service
class MetricsService(
    private val repositoryService: RepositoryService,
    private val toolRunService: ToolRunService,
    private val repositoryDetailsService: RepositoryDetailsService,
    private val kpiService: KPIService,
    private val ortService: OrtService,
    private val occmdService: OccmdService
) {

    private val defaultScope = CoroutineScope(Dispatchers.Default)
    val ioScope = CoroutineScope(Dispatchers.IO)


    /**
     * Warning: Long running function!
     *
     * In this function we calculate the KPIs for the given project using information from the gitlab api.
     * Therefore, we first create a repository and tool run object in the DB.
     *
     * Next, we asynchronously query the gitlab API and dedicated opencode tool APIs for further data.
     * Based upon this data we calculate RAW KPIs.
     *
     * Once all API responses have been processes we create a KPI tree and store this to the DB.
     *
     * Lastly, we update the tool run with all tools, which provided results for the given repository,
     * and store this to the DB.
     */
    suspend fun handleRepositoryChange(repoId: Long) = defaultScope.launch {

        /**
         * Create initial db models.
         * If this fails we want the whole method to fail as we can't continue
         * without the db models.
         *
         */
        val toolRun = async(ioScope.coroutineContext) {
            val repoDto = repositoryService.getRepositoryInfo(repoId)
            val repo = repositoryService.getOrCreate(repoDto)
            toolRunService.createToolRunForRepository(repo)
        }.await()

        /**
         * For all tools, first query the tool results API.
         * If the tool has results, add the tool to the tool run
         * and then calculate the KPI based on the results, link
         * the KPI with the repository and store everything in the db.
         * The function returns the UUID of the KPI in the db.
         * If we can calculate multiple KPIs from one tool we can return
         * a list of UUIDs.
         *
         * val toolKpi = async {
         *  val toolResult = queryToolApi(repoId)
         *  if (!toolResult.isEmpty) {
         *      toolRun.addTool(tool)
         *      val kpi = calculateKpi(toolResults)
         *      kpi.addRepository(toolRun.repo)
         *      kpi.store()
         *      kpi.id // this is the return of the async function
         *  }
         * }
         */

        val vulnerabilityKpis = async {
            val vulnerabilityDtos = ortService.getOrtResults(repoId) // in the dev setup we get results for repo id 106
            val kpis = ortService.calculateVulnerabilityKpis(vulnerabilityDtos)
            if (kpis.isNotEmpty()) {
                toolRun.toolEntities.add(ortService.toolEntity)
            }
            kpis
        }

        val repositoryDetailsKpi = async {
            val repoDetailsDto = repositoryDetailsService.getRepositoryDetails(repoId)
            toolRun.toolEntities.add(repositoryDetailsService.toolEntity)
            repositoryDetailsService.calculateRepositoryDetailsKpis(repoDetailsDto)
        }

        val occmdKpis = async {
            val rawOccmdResults = occmdService.runOccmd(repoId, toolRun.repository?.url!!)
            if (rawOccmdResults.isNotEmpty()) {
                toolRun.toolEntities.add(occmdService.toolEntity)
            }
            occmdService.calculateOccmdKpis(rawOccmdResults)
        }

        /**
         * Calculate KPI tree by awaiting all tool KPIs
         * joinAll(toolKpi)
         * calculateKpiTree(toolKpi, ...)
         */
        kpiService.calculateAndSaveKpiTree(
            repository = toolRun.repository!!,
            repositoryDetailsKpi = repositoryDetailsKpi.await(),
            vulnerabilityKpis = vulnerabilityKpis.await(),
            occmdKpis = occmdKpis.await()
        )

        toolRunService.saveToolRunEntitiy(toolRun)
    }

}
